/* df3util.h */

#ifndef DF3UTIL_H
#define DF3UTIL_H 20170501

/* cf PATH_MAX, strndup(3), etc */
#define _POSIX_C_SOURCE 200809L

#include <df3.h>

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>


/* status/error codes */
enum Ecodes {
  ec_ok = 0,
  ec_badBPV,
  ec_badAxes,
  ec_badOffset,
  ec_badVoxel,
  ec_badArgument,
  ec_diskFull,
  ec_badFilename,
  ec_existFile,
  ec_permsFile,
  ec_noFile,
  ec_badUsage,
  ec_sysError,
  ec_ec_max
};
typedef enum Ecodes Ecodes;


/* stdin formats */
enum inpTypes {
  itNone = 0,
  itCsv,
  itTxt,
  itTcl
};
typedef enum inpTypes inpTypes;

/* generic origin */
#define GENORIGIN {0,0,0}

/* offsets shift the origins in volumes,
 * legal range -65533,..,65534.
 */
typedef struct Offset Offset;
struct Offset { int x, y, z; };

#define ZEROOFFS (Offset) GENORIGIN


/* common.c */
bool mkSourceFname(char const * const, bool const, bool const, char **);
bool mkTargetFname(char const * const, bool const, char **);
void mkBufferXY(df3Axes const, char const, size_t *, size_t *);
size_t axisMax(df3Axes const, char const);
uint32_t getFillValue(size_t const);
int getInpSrc( char *[], df3Axes *, size_t *);
bool isDefaultArg(char const * const);
bool isValidFillval(size_t const);
bool isValidVoxel(size_t const, uint32_t const);
bool inpVoxelValid(df3Axes const, size_t const, df3Coord const, uint32_t const);
df3Coord atOffset(bool const, df3Coord const, Offset const);
bool badAcol(df3Axes const, char const, size_t const);
char const *trimmedFn(char const * const);
bool prepFilled(char const * const, df3Handle *);


/* vinters.c */
bool containedIn(df3Axes const, Offset const, df3Coord const);
bool volsOverlap(df3Axes const, Offset const, df3Coord const, bool *);
bool atCGTinters(df3Axes const, df3Coord *, Offset const, df3Coord const);
bool atCLEinters(df3Axes const, df3Coord *, Offset const, df3Coord const);
bool atMinters(df3Axes const, df3Coord *, Offset const, df3Coord const);



/* errors.c */
char const *appLastError(Ecodes const);
Ecodes le2ae(df3Error const);
void appPanic(char const * const, int const) __attribute__((noreturn));
#define APPPANIC  appPanic(__FILE__,__LINE__)


/* pngrw.c */
bool pngGet(char const * const, size_t const, size_t const, size_t const, void *);
bool pngPut(char const * const, size_t const, size_t const, size_t const, void *);


/* regex.c */
bool getArgAxisN(char const * const, char *, int *);
bool getArgCoord(char const * const , df3Coord *);
bool getArgFname(char const * const, char *);
bool getArgNumber(char const * const, int *);
bool getArgOffset(char const * const, Offset *);
bool getArgRange(char const * const, uint32_t *, uint32_t *);
bool getArgValue(char const * const, uint32_t *);
bool getArgVoxel(char const * const, bool const, df3Coord *, uint32_t *);
void setupRE(void);


/* inopt.c */
bool inputGet(FILE *, bool (*)(void *, df3Coord const, uint32_t const), void *);


/* stat.c */
void setupGrp(void);
bool haveDiskSpace(df3Axes const, size_t const);
bool statDirectoryPerms(char const * const);
bool statFileExists(char const * const);
bool statFilePerms(char const * const, bool *);


/* main.c */
extern bool app_diagnose;
extern long int app_fillval;
extern double app_gamma;
extern inpTypes app_inptype;
extern Ecodes app_lastec;
extern FILE *app_logfp;
extern char *app_mapfile;
extern char const app_name[];
extern bool app_verbose;
extern char const app_version[];
extern bool app_warns;
extern char *app_workdir;


/* abbrev */
#define AUU  __attribute__((unused))
/* chars */
#define CHRDOT  '.'
#define CHRHYP  '-'
#define CHRNUL  '\0'
#define CHROCT  '#'
#define PTHSEP  '/'
/* same as WindowMaker */
#define ENVVAR  "DISPLAY_GAMMA"
/* message buffer size */
#define MBUFSZ  1024
/* allow for filename */
#define PTHNET  (PATH_MAX - 20)

#define EMITDMSG(tu_,...)               \
do {                                    \
  char buf_[MBUFSZ];                    \
  snprintf(buf_, MBUFSZ, __VA_ARGS__);  \
  fprintf(app_logfp, "DIAG:%s: %s.\n",  \
          (tu_), (buf_));               \
} while (0)

#define EMITEMSG(w_,...)                 \
do {                                     \
  char buf_[MBUFSZ];                     \
  snprintf(buf_, MBUFSZ, __VA_ARGS__);   \
  if (stderr != app_logfp)               \
    fprintf(app_logfp, ": %s : %s.\n",   \
            (w_) ? "warning" : "error",  \
            buf_);                       \
  fprintf(stderr, "%s: %s %s.\n",        \
          app_name,                      \
          (w_) ? "warning" : "error",    \
          buf_);                         \
} while (0)

#define FREENULL(p_)  \
do {                  \
  free(p_);           \
  (p_) = NULL;        \
} while (0)


#endif

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
