/* inopt.c */

#include "df3util.h"

#include <ctype.h>


#define CHRCOM  ','
#define CHRNLN  '\n'
#define CHRSPC  ' '

#define CSVLNMAX  512
#define TXTLNMAX  40
#define VOXLMAX   28

#define BADREAD  "input stream error on reading"
#define FMTLINE  "bad data in line %d"
#define FMTVOXL  "bad/incomplete data for voxel #%ld"


/* doTcl()
 * -------------------------------------------------------------------------- *
 * expect flat list of one or more voxels.
 * note for now accept Tcl lists with missing elements.
 */

typedef enum scState {
  scNumEnd,
  scNumBegin,
  scNumCont
} scState;

static bool buildLn(char *ln, scState const sc, int const ch, bool *done) {
  static uint64_t cnt = 0;
  size_t const n = strlen(ln);
  if (VOXLMAX < n)
    return false;
  switch (sc) {
    case scNumBegin:  /* fall through */
      *done = false;
    case scNumCont:
      ln[n] = (char)ch;
      break;
    case scNumEnd:
      cnt += 1;
      if (cnt % 4)
        ln[n] = CHRSPC;
      else {
        ln[n] = CHRNLN;
        *done = true;
      }
      break;
  }
  return true;
}

static bool doTcl(FILE *fp,
                  bool (*callback)(void *, df3Coord const, uint32_t const),
                  void *cdata) {
  bool currdone = true,  /* have xyzv? */
       f = false,
       sspc = true;  /* (state) space between numbers */
  char currln[TXTLNMAX+1] = {CHRNUL};
  int ch;
  long int vnum = 1;

  while (EOF != (ch = fgetc(fp)))
    if (sspc) {
      if (isdigit(ch)) {
        if (!buildLn(currln, scNumBegin, ch, &currdone)) {
          EMITEMSG(false, FMTVOXL, vnum);
          app_lastec = ec_badVoxel;
          return false;
        }
        sspc = false;
      }
      else if (!isspace(ch)) {
        EMITEMSG(false, FMTVOXL, vnum);
        app_lastec = ec_badVoxel;
        return false;
      }
      /* for now ignore consecutive spaces */
    }
    else {
      if (isdigit(ch)) {
        if (!buildLn(currln, scNumCont, ch, &currdone)) {
          EMITEMSG(false, FMTVOXL, vnum);
          app_lastec = ec_badVoxel;
          return false;
        }
      }
      else if (!isspace(ch)) {
        EMITEMSG(false, FMTVOXL, vnum);
        app_lastec = ec_badVoxel;
        return false;
      }
      else {
        if (!buildLn(currln, scNumEnd, CHRNUL, &currdone)) {
          EMITEMSG(false, FMTVOXL, vnum);
          app_lastec = ec_badVoxel;
          return false;
        }
        else if (currdone) {
          df3Coord at;
          uint32_t val;
          if (!getArgVoxel(currln, false, &at, &val)) {
            EMITEMSG(false, FMTVOXL, vnum);
            return false;
          }
          else if (!callback(cdata, at, val))
            return false;
          memset(currln, 0, TXTLNMAX);
          vnum += 1;
        }
        sspc = true;
      }
    }

  if (ferror(fp)) {
    EMITEMSG(false, BADREAD);
    app_lastec = ec_sysError;
  }
  else if (!currdone) {
    EMITEMSG(false, FMTVOXL, vnum);
    app_lastec = ec_badVoxel;
  }
  else
    f = true;

  return f;
}


/* doTxt()
 * -------------------------------------------------------------------------- *
 * space-separated values.  ignore empty lines + comments.
 * lines may be up to 40 chars long, though 29 (incl NL) is the longest
 * legal line.
 */

static bool doTxt(FILE *fp,
                  bool (*callback)(void *, df3Coord const, uint32_t const),
                  void *cdata) {
  bool f = false;
  char line[TXTLNMAX+1] = {CHRNUL};
  int lnnum = 0;

  while (fgets(line, TXTLNMAX, fp)) {
    df3Coord at;
    uint32_t val;
    lnnum += 1;
    if (CHROCT == *line  ||  CHRNLN == *line)
      continue;
    else  if (!getArgVoxel(line, false, &at, &val)) {
      EMITEMSG(false, FMTLINE, lnnum);
      return false;
    }
    else if (!callback(cdata, at, val))
      return false;
    memset(line, 0, TXTLNMAX);
  }

  if (ferror(fp)) {
    EMITEMSG(false, BADREAD);
    app_lastec = ec_sysError;
  }
  else
    f = true;

  return f;
}


/* doCsv()
 * -------------------------------------------------------------------------- *
 * comma-separated values.  ignore empty lines + comments.
 * lines may be up to 512 chars long, more than enough for 16 voxels.
 * conditioning means line must have trailing newline, and no comma before.
 * if there is more than one voxel in the line, convert every fourth comma
 * to a newline, comma count always multiple of three; results in null-
 * terminated list of newline separated strings ready (bar NL replacing)
 * for getArg*() conversion.
 */

static bool conditionLine(char * const ln) {
  char *p = ln, *x;
  int n = 0;
label_again:
  while (3 != n  &&  (p = strchr(p + 1, CHRCOM)))
    n += 1;
  if (!p  ||  3 != n)
    return false;
  else if ((x = strchr(p + 1, CHRCOM))) {
    *x = CHRNLN;
    p = x;
    n = 0;
    goto label_again;
  }
  else if (!(x = strchr(p + 1, CHRNLN)))
    return false;
  return true;
}

static bool doCsv(FILE *fp,
                  bool (*callback)(void *, df3Coord const, uint32_t const),
                  void *cdata) {
  bool f = false;
  char line[CSVLNMAX+1] = {CHRNUL};
  int lnnum = 0;

  while (fgets(line, CSVLNMAX, fp)) {
    lnnum += 1;
    if (CHROCT == *line  ||  CHRNLN == *line)
      continue;
    else if (!conditionLine(line)) {
      EMITEMSG(false, FMTLINE, lnnum);
      app_lastec = ec_badVoxel;
      return false;
    }
    char *p = line;
    while (*p) {
      df3Coord at;
      uint32_t val;
      char *x = strchr(p, CHRNLN);
      if (!x)
        APPPANIC;
      *x = CHRNUL;
      if (!getArgVoxel(p, true, &at, &val)) {
        EMITEMSG(false, FMTLINE, lnnum);
        return false;
      }
      else if (!callback(cdata, at, val))
        return false;
      p = x + 1;
    }
    memset(line, 0, CSVLNMAX);
  }

  if (ferror(fp)) {
    EMITEMSG(false, BADREAD);
    app_lastec = ec_sysError;
  }
  else
    f = true;

  return f;
}


/* inputGet()
 * -------------------------------------------------------------------------- *
 */

bool inputGet(FILE *fp,
              bool (*callback)(void *, df3Coord const, uint32_t const),
              void *cdata) {

  if (!fp)
    fp = stdin;

  if (ferror(fp)) {
    EMITEMSG(false, "input stream error before first read");
    app_lastec = ec_sysError;
    return false;
  }

  bool f = false;
  switch (app_inptype) {
    case itCsv: f = doCsv(fp, callback, cdata); break;
    case itTxt: f = doTxt(fp, callback, cdata); break;
    case itTcl: f = doTcl(fp, callback, cdata); break;
    default:    APPPANIC;
  }
  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
