/* mirror.c */

#include "df3util.h"

static char const * const TU = "mirror";

/* globals */
static char *asrc,      /* source name arg, null if default */
            *atgt;      /* target name arg */
static char aaxis;      /* axis arg */
static size_t apivot;   /* pivot column arg */

static df3Axes saxes;   /* source axes arg, if source default */
static size_t sbpv;     /* source bpv arg, if source default */

static df3Handle df3i, df3o;

/* mirroring */
static size_t pivotPoint(size_t const axlen,
                         size_t const pivot,
                         size_t const pt) {
  int const ilen = (int)axlen, ipiv = (int)pivot, ipt = (int)pt + 1;
  size_t res;
  if (1 == axlen)
    res = pt;
  else if (pivot <= pt)
    res = (size_t)((ilen + ipiv - (ipt - ipiv)) % ilen);
  else
    res = (size_t)((ilen + (ipiv - ipt) + ipiv) % ilen);
  return res;
}

static void mirrorAt(df3Axes const axes,
                     char const axis,
                     size_t const pivot,
                     df3Coord *at) {
  switch (axis) {
    case 'x':  at->x = pivotPoint(axes.x, pivot, at->x);  break;
    case 'y':  at->y = pivotPoint(axes.y, pivot, at->y);  break;
    case 'z':  at->z = pivotPoint(axes.z, pivot, at->z);  break;
    default:   APPPANIC;
  }
  return;
}


/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := source axis:n target.
 *     := . axes bpv axis:n target.
 */

static bool procArgs(char *av[]) {
  bool f = false;
  char buf[PATH_MAX];
  int const rc = getInpSrc(av, &saxes, &sbpv);
  int i;

  if (-1 == rc)
    goto label_fail;
  else if (1 == rc)
    i = 3;
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    goto label_fail;
  }
  else {
    asrc = av[0];
    i = 1;
  }

  if (!(av[i]  &&  av[i+1])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    goto label_fail;
  }

  int n = 0;  /* default if omitted */
  if (!getArgAxisN(av[i], &aaxis, &n)) {
    EMITEMSG(false, "bad axis:n argument");
    goto label_fail;
  }
  else if (0 > n) {
    EMITEMSG(false, "bad axis:n argument");
    app_lastec = ec_badUsage;
    goto label_fail;
  }
  apivot = (size_t)n;

  if (!getArgFname(av[i+1], buf)) {
    EMITEMSG(false, "bad target file name");
    goto label_fail;
  }
  atgt = av[i+1];

  f = true;

label_fail:
  return f;
}


/* openSource()
 * -------------------------------------------------------------------------- *
 * open if actual file, check pivot fits.
 */

static bool badPivot(df3Axes const axes, char const axis, size_t const n) {
  bool f = false;
  switch (axis) {
    case 'x':  f = axes.x <= n;  break;
    case 'y':  f = axes.y <= n;  break;
    case 'z':  f = axes.z <= n;  break;
    default:   APPPANIC;
  }
  return f;
}

static bool openSource(void) {

  if (asrc) {
    df3Error ec;
    if (!mkSourceFname(asrc, false, false, &df3i.fname)) {
      EMITEMSG(false, "source '%s': %s",
              trimmedFn(asrc), appLastError(app_lastec));
      return false;
    }
    else if (ecOk != (ec = df3_open_file(&df3i, true))) {
      EMITEMSG(false, "open source '%s': %s",
              trimmedFn(df3i.fname), df3_error_message(ec));
      app_lastec = le2ae(ec);
      return false;
    }
  }
  else {
    df3i.axes = saxes;
    df3i.bpv = sbpv;
  }

  if (badPivot(df3i.axes, aaxis, apivot)) {
    EMITEMSG(false, "axis '%c' range 0,..,%zu",
            aaxis, axisMax(df3i.axes, aaxis) - 1);
    app_lastec = ec_badArgument;
    return false;
  }

  if (app_diagnose) {
    if (asrc) {
      EMITDMSG(TU, "source handle");
      df3_handle_info(&df3i, app_logfp);
    }
    else
      EMITDMSG(TU, "source handle <stdin>");
  }

  return true;
}


/* openTarget()
 * -------------------------------------------------------------------------- *
 * create file, fill if source stdin.
 */

static bool openTarget(void) {

  if (!mkTargetFname(atgt, false, &df3o.fname)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(atgt), appLastError(app_lastec));
    return false;
  }
  else if (!haveDiskSpace(df3i.axes, df3i.bpv)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(df3o.fname), appLastError(app_lastec = ec_diskFull));
    return false;
  }

  df3o.axes = df3i.axes;
  df3o.bpv = df3i.bpv;

  df3Error ec = df3_create_file(&df3o);
  if (ecOk != ec) {
    EMITEMSG(false, "create target '%s': %s",
            trimmedFn(df3i.fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3o, app_logfp);
  }

  if (!asrc)
    if (!prepFilled(TU, &df3o)) {
      (void)df3_close_file(&df3o);
      return false;
    }

  return true;
}


/* closeFiles()
 * -------------------------------------------------------------------------- *
 * TODO what's lost by setting {app_lastec} here?
 * TODO use EMITEMSGs instead + unconditionally?
 */

static bool closeFiles(void) {
  bool f = false;
  df3Error ec;

  if (asrc) {
    if (ecOk != (ec = df3_close_file(&df3i))) {
      if (app_diagnose)
        EMITDMSG(TU, "close source: %s", df3_error_message(ec));
    }
    else if (app_diagnose)
      EMITDMSG(TU, "closed source handle: %s",
              df3_handle_error(&df3i, false));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {  /* redundant, but hey.. */
    (void)fclose(stdin);
    if (app_diagnose)
      EMITDMSG(TU, "closed source <stdin>");
  }

  ec = df3_close_file(&df3o);
  if (ecOk != ec) { 
    if (app_diagnose)
      EMITDMSG(TU, "close target: %s", df3_error_message(ec));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {
    if (app_diagnose)
      EMITDMSG(TU, "closed target handle: %s", df3_handle_error(&df3o, false));
    f = true;
  }

  return f;
}


/* procInp()
 * -------------------------------------------------------------------------- *
 * wraps callback via inputGet().
 */

static bool cback(void *vp, df3Coord const at, uint32_t const v) {
  df3Error ec;
  df3Handle * const df3 = vp;
  df3Coord pos = at;

  if (!inpVoxelValid(df3->axes, df3->bpv, at, v))
    return false;

  mirrorAt(df3->axes, aaxis, apivot, &pos);

  if (ecOk != (ec = df3_put_voxel(df3, pos, v))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            pos.x, pos.y, pos.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  return true;
}

static inline bool procInp(void) {
  return inputGet(NULL, cback, &df3o);
}


/* procFile()
 * -------------------------------------------------------------------------- *
 */

static bool procFile(void) {
  bool f = false;

  for (size_t z = 0; df3i.axes.z != z; ++z)
    for (size_t y = 0; df3i.axes.y != y; ++y)
      for (size_t x = 0; df3i.axes.x != x; ++x) {
        df3Error ec;
        uint32_t v;
        df3Coord at = {x, y, z};
        if (ecOk != (ec = df3_get_voxel(&df3i, at, &v))) {
          EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        mirrorAt(df3o.axes, aaxis, apivot, &at);
        if (ecOk != (ec = df3_put_voxel(&df3o, at, v))) {
          EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }

  f = true;

label_loopfail:
  return f;
}


/* ------------------------------------------------------------------------- */

static void mirrorTidyup(void) {
  FREENULL(df3i.fname);
  FREENULL(df3o.fname);
  return;
}

bool mirrorMain(char *av[]) {
  if (atexit(mirrorTidyup))
    APPPANIC;

  if (!procArgs(av))
    return false;
  else if (!openSource())
    return false;
  else if (!openTarget()) {
    (void)df3_close_file(&df3i);
    return false;
  }

  bool const f = (asrc ? procFile : procInp)();

  if (!(closeFiles()  &&  f)) {
    (void)unlink(df3o.fname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(df3o.fname));
  }
  else if (app_verbose)
    printf("volume mirrored along '%c:%zu'.\n", aaxis, apivot);

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
