/* recode.c */

/* cf tdestroy(3) */
#define _GNU_SOURCE

#include "df3util.h"

#include <search.h>

static char const * const TU = "recode";

/* globals */
static char *asrc,        /* source name arg, null if default */
            *atgt;        /* target name arg */
static size_t abpv;       /* n arg */

static df3Axes saxes;     /* source axes arg, if source default */
static size_t sbpv;       /* source bpv arg, if source default */

static size_t avidx;      /* index of first {from} if any */

static df3Handle df3i, df3o;

static uint32_t ntuples,  /* counters */
                nmapped;
static uint64_t nvcopy,
                nvmap,
                nvtrunc;

enum Maptype {            /* infrastucture */
  mtInvalid = 0,
  mtV2V,
  mtV2R,
  mtR2V,
  mtR2R
};
typedef enum Maptype Maptype;

typedef struct Tuple Tuple;
struct Tuple {
  uint32_t from[2];
  uint32_t to[2];
  Maptype typ;
};

static void *tuples;

static int cmpTuples(void const *vp1, void const *vp2) {
  Tuple const *tp1 = vp1, *tp2 = vp2;
  bool const range1 = (mtR2V == tp1->typ  ||  mtR2R == tp1->typ),
             range2 = (mtR2V == tp2->typ  ||  mtR2R == tp2->typ);
  if ((range1 ? tp1->from[1] : tp1->from[0]) < tp2->from[0])
    return -1;
  else if ((range2 ? tp2->from[1] : tp2->from[0]) < tp1->from[0])
    return 1;
  else
    return 0;
}

static bool applyMap(size_t const sb,
                     size_t const tb,
                     df3Coord const at,
                     uint32_t *pval) {
  bool f = true;
  uint32_t v = *pval;
  Tuple cmp = {{*pval,0},{0,0}, mtV2V};
  void *tmp = tfind(&cmp, &tuples, cmpTuples);

  if (tmp) {
    Tuple *p = *(Tuple **)tmp;
    switch (p->typ) {
      case mtV2V:
      case mtR2V:
        v = p->to[0];
        break;
      case mtV2R:
        v = p->to[1] - p->to[0] + 1;
        v = (uint32_t)random() % v;
        v = p->to[0] + v;
        break;
      case mtR2R:
        v = v - p->from[0];
        v = v + p->to[0];
        break;
      default:
        APPPANIC;
    }
    nvmap += 1;
  }
  else if (sb <= tb)
    nvcopy += 1;
  else if (isValidVoxel(tb, v))
    nvcopy += 1;
  else if (!app_warns) {
    EMITEMSG(false, "would truncate voxel <%zu,%zu,%zu>", at.x, at.y, at.z);
    app_lastec = ec_badVoxel;
    f = false;
  }
  else {
    v = (2 == tb ? UINT16_MAX : UINT8_MAX) & v;
    nvtrunc += 1;
    if (app_diagnose)
      EMITEMSG(true, "truncated voxel <%zu,%zu,%zu>", at.x, at.y, at.z);
  }

  *pval = v;

  return f;
}


/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := source bpv target [from to].
 *     := . axes bpv bpv target [from to].
 */

static bool procArgs(char *av[]) {
  bool f = false;
  char buf[PATH_MAX];
  int const rc = getInpSrc(av, &saxes, &sbpv);
  int i;

  if (-1 == rc)
    goto label_fail;
  else if (1 == rc)
    i = 3;
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    goto label_fail;
  }
  else {
    asrc = av[0];
    i = 1;
  }

  if (!(av[i]  &&  av[i+1])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    goto label_fail;
  }

  uint32_t b;
  if (!getArgValue(av[i], &b)) {
    EMITEMSG(false, "bad bpv value");
    return false;
  }
  else if (!df3_check_bpv((size_t)b)) {
    EMITEMSG(false, "bad bpv value");
    app_lastec = ec_badBPV;
    return false;
  }
  abpv = (size_t)b;


  if (!getArgFname(av[i+1], buf)) {
    EMITEMSG(false, "bad target file name");
    goto label_fail;
  }
  atgt = av[i+1];

  avidx = (size_t)i + 2;

  f = true;

label_fail:
  return f;
}


/* openSource()
 * -------------------------------------------------------------------------- *
 * open if actual file, pre-compute addend.
 */


static bool openSource(void) {

  if (asrc) {
    df3Error ec;
    if (!mkSourceFname(asrc, false, false, &df3i.fname)) {
      EMITEMSG(false, "source '%s': %s",
              trimmedFn(asrc), appLastError(app_lastec));
      return false;
    }
    else if (ecOk != (ec = df3_open_file(&df3i, true))) {
      EMITEMSG(false, "open source '%s': %s",
              trimmedFn(df3i.fname), df3_error_message(ec));
      app_lastec = le2ae(ec);
      return false;
    }
  }
  else {
    df3i.axes = saxes;
    df3i.bpv = sbpv;
  }

// TODO check bpv combination cf app_warns.  diag feedback.

  if (app_diagnose) {
    if (asrc) {
      EMITDMSG(TU, "source handle");
      df3_handle_info(&df3i, app_logfp);
    }
    else
      EMITDMSG(TU, "source handle <stdin>");
  }

  return true;
}


/* openTarget()
 * -------------------------------------------------------------------------- *
 * create file, fill if source stdin.
 */

static bool openTarget(void) {

  if (!mkTargetFname(atgt, false, &df3o.fname)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(atgt), appLastError(app_lastec));
    return false;
  }
  else if (!haveDiskSpace(df3i.axes, abpv)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(df3o.fname), appLastError(app_lastec = ec_diskFull));
    return false;
  }

  df3o.axes = df3i.axes;
  df3o.bpv = abpv;

  df3Error ec = df3_create_file(&df3o);
  if (ecOk != ec) {
    EMITEMSG(false, "create target '%s': %s",
            trimmedFn(df3i.fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3o, app_logfp);
  }

  if (!asrc)
    if (!prepFilled(TU, &df3o)) {
      (void)df3_close_file(&df3o);
      return false;
    }

  return true;
}


/* closeFiles()
 * -------------------------------------------------------------------------- *
 * TODO what's lost by setting {app_lastec} here?
 * TODO use EMITEMSGs instead + unconditionally?
 */

static bool closeFiles(void) {
  bool f = false;
  df3Error ec;

  if (asrc) {
    if (ecOk != (ec = df3_close_file(&df3i))) {
      if (app_diagnose)
        EMITDMSG(TU, "close source: %s", df3_error_message(ec));
    }
    else if (app_diagnose)
      EMITDMSG(TU, "closed source handle: %s",
              df3_handle_error(&df3i, false));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {  /* redundant, but hey.. */
    (void)fclose(stdin);
    if (app_diagnose)
      EMITDMSG(TU, "closed source <stdin>");
  }

  ec = df3_close_file(&df3o);
  if (ecOk != ec) { 
    if (app_diagnose)
      EMITDMSG(TU, "close target: %s", df3_error_message(ec));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {
    if (app_diagnose)
      EMITDMSG(TU, "closed target handle: %s", df3_handle_error(&df3o, false));
    f = true;
  }

  return f;
}


/* procInp()
 * -------------------------------------------------------------------------- *
 * wraps callback via inputGet().
 */

static bool cback(void *vp, df3Coord const at, uint32_t const v) {
  df3Error ec;
  df3Handle * const df3 = vp;
  uint32_t nv = v;

  if (!inpVoxelValid(df3->axes, df3->bpv, at, v))
    return false;
  else if (!applyMap(df3i.bpv, df3->bpv, at, &nv))
    return false;
  else if (ecOk != (ec = df3_put_voxel(df3, at, nv))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            at.x, at.y, at.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  return true;
}

static inline bool procInp(void) {
  return inputGet(NULL, cback, &df3o);
}


/* procFile()
 * -------------------------------------------------------------------------- *
 */

static bool procFile(void) {
  bool f = false;

  for (size_t z = 0; df3i.axes.z != z; ++z)
    for (size_t y = 0; df3i.axes.y != y; ++y)
      for (size_t x = 0; df3i.axes.x != x; ++x) {
        df3Error ec;
        uint32_t v;
        df3Coord at = {x, y, z};
        if (ecOk != (ec = df3_get_voxel(&df3i, at, &v))) {
          EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        else if (!applyMap(df3i.bpv, df3o.bpv, at, &v))
          goto label_loopfail;
        else if (ecOk != (ec = df3_put_voxel(&df3o, at, v))) {
          EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }

  f = true;

label_loopfail:
  return f;
}


/* procMappings()
 * -------------------------------------------------------------------------- *
 * process -mapfile, if any, then command-line args, if any.
 * false on error: a bad command line, an invalid mapping (bpv or overlap),
 * or file i/o problems.  true on success.
 * mapfile already checked exist/regular/readable.
 * NB (too many) magic numbers/literals.
 */

static bool addMapTuple(size_t const sb,
                        size_t const tb,
                        char const * const from,
                        char const * const to) {
  bool ffr, ftr;
  Tuple tmp = {{0,0},{0,0},mtInvalid};

  ntuples += 1;

  if (!(ffr = getArgRange(from, &tmp.from[0], &tmp.from[1])))
    if (!getArgValue(from, &tmp.from[0]))
      return false;

  if (ffr  &&  tmp.from[1] <= tmp.from[0]) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (!isValidVoxel(sb, tmp.from[ffr ? 1 : 0])) {
    app_lastec = ec_badArgument;
    return false;
  }

  if (!(ftr = getArgRange(to, &tmp.to[0], &tmp.to[1])))
    if (!getArgValue(to, &tmp.to[0]))
      return false;

  if (ftr  &&  tmp.to[1] <= tmp.to[0]) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (!isValidVoxel(tb, tmp.to[ftr ? 1 : 0])) {
    app_lastec = ec_badArgument;
    return false;
  }

  if (!(ffr || ftr))
    tmp.typ = mtV2V;
  else if (!ffr  &&  ftr)
    tmp.typ = mtV2R;
  else if (ffr  &&  !ftr)
    tmp.typ = mtR2V;
  else {
    uint32_t const d1 = tmp.from[1] - tmp.from[0],
                   d2 = tmp.to[1] - tmp.to[0];
    if (d1 != d2) {
      app_lastec = ec_badArgument;
      return false;
    }
    tmp.typ = mtR2R;
  }

  Tuple *p;
  if (tfind(&tmp, &tuples, cmpTuples)) {
// TODO ec overlapping?
    app_lastec = ec_badArgument;
    return false;
  }
  else if (!(p = malloc(sizeof(Tuple))))
    APPPANIC;
  *p = tmp;
  if (!tsearch(p, &tuples, cmpTuples))
    APPPANIC;

  if (mtV2V == tmp.typ  ||  mtV2R == tmp.typ)
    nmapped += 1;
  else
    nmapped += tmp.from[1] - tmp.from[0] + 1;

  return true;
}


static bool procMapFile(size_t const from, size_t const to) {
  FILE *fp = fopen(app_mapfile, "r");
  if (!fp) {
    int const e = errno;
    EMITEMSG(false, "opening map file '%s': %s",
            trimmedFn(app_mapfile), strerror(e));
    app_lastec = ec_sysError;
    return false;
  }
  else if (app_diagnose)
    EMITDMSG(TU, "reading map file '%s'", trimmedFn(app_mapfile));

// TODO app_lastec.
  bool f = false;
  char line[MBUFSZ];
  int ln = 0;
  size_t bpv1, bpv2;
  while (fgets(line, MBUFSZ, fp)) {
    ln += 1;
    if (!strncmp("map ", line, 4))
      break;
  }

  if (feof(fp)  ||  ferror(fp)) {
    EMITEMSG(false, "unexpected EOF");
    goto label_fail;
  }
  else if (2 != sscanf(line, "map %zu %zu", &bpv1, &bpv2)) {
    EMITEMSG(false, "bad map spec in '%s' line %d",
            trimmedFn(app_mapfile), ln);
    goto label_fail;
  }
  else if (!df3_check_bpv(bpv1)  ||  bpv1 != from) {
    EMITEMSG(false, "bad source map bpv in '%s' line %d",
            trimmedFn(app_mapfile), ln);
    goto label_fail;
  }
  else if (!df3_check_bpv(bpv2)  ||  bpv2 != to) {
    EMITEMSG(false, "bad target map bpv in '%s' line %d",
            trimmedFn(app_mapfile), ln);
    goto label_fail;
  }

  char arg1[22], arg2[22];
  int nread = 0;
  while (fgets(line, MBUFSZ, fp)) {
    ln += 1;
    if (CHROCT == line[0]  ||  ';' == line[0]  ||  2 > strlen(line))
      continue;
    else if (2 != sscanf(line, "%21s %21s", arg1, arg2)) {
      EMITEMSG(false, "bad map value(s) in '%s' line %d",
              trimmedFn(app_mapfile), ln);
      goto label_fail;
    }
    else if (!addMapTuple(from, to, arg1, arg2)) {
      EMITEMSG(false, "bad map value(s) in '%s' line %d: %s",
              trimmedFn(app_mapfile), ln, appLastError(app_lastec));
      goto label_fail;
    }
    else
      nread += 1;
  }

  if (!nread) {
    EMITEMSG(false, "no map data in file");
    goto label_fail;
  }
  else if (app_diagnose)
    EMITDMSG(TU, "read %d map tuple%s from file",
            nread, (1 == nread) ? "" : "s");

  f = true;

label_fail:
  (void)fclose(fp);

  return f;
}


static bool procMappings(size_t const from, size_t const to, char *av[]) {
  if (app_mapfile)
    if (!procMapFile(from, to))
      return false;

  bool f = true;
  int n = 1;
  for (size_t i = 0; av[i]; i += 2, ++n)
    if (!av[i+1]) {
      EMITEMSG(false, "bad map tuple #%d", n);
      app_lastec = ec_badArgument;
      f = false;
      break;
    }
    else if (!addMapTuple(from, to, av[i], av[i+1])) {
      EMITEMSG(false, "bad map tuple #%d: %s", n, appLastError(app_lastec));
      f = false;
      break;
    }

  if (f)
    if (app_diagnose) {
      n -= 1;
      EMITDMSG(TU, "read %d map tuple%s from command line",
              n, (1 == n) ? "" : "s");
      if (ntuples) {
        double const mn = 1.0 + (4 == from ?
                UINT32_MAX : (double)(2 == from ? UINT16_MAX : UINT8_MAX));
        EMITDMSG(TU, "%u tuple%s, mapping %u value%s (%.2f%% coverage)",
                ntuples, 1 == ntuples ? "" : "s",
                nmapped, 1 == nmapped ? "" : "s",
                (double)nmapped / (mn / 100));
      }
    }

  return f;
}


/* ------------------------------------------------------------------------- */

static void recodeTidyup(void) {
  if (tuples) {
    tdestroy(tuples, free);
    tuples = NULL;
  }
  FREENULL(df3i.fname);
  FREENULL(df3o.fname);
  return;
}

bool recodeMain(char *av[]) {
  if (atexit(recodeTidyup))
    APPPANIC;

  if (!procArgs(av))
    return false;
  else if (!openSource())
    return false;
  else if (!procMappings(df3i.bpv, abpv, av + avidx)) {
    (void)df3_close_file(&df3i);
    return false;
  }
  else if (!openTarget()) {
    (void)df3_close_file(&df3i);
    return false;
  }

  bool const f = (asrc ? procFile : procInp)();

  if (!(closeFiles()  &&  f)) {
    (void)unlink(df3o.fname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(df3o.fname));
  }
  else if (app_verbose)
    printf("%lu voxel%s copied, %lu voxel%s mapped, %lu voxel%s truncated.\n",
           nvcopy, 1 == nvcopy ? "" : "s",
           nvmap, 1 == nvmap ? "" : "s",
           nvtrunc, 1 == nvtrunc ? "" : "s");

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
